@@ -1,4 +1,6 @@
 from __future__ import print_function
+from functools import wraps
+from functools import partial
 
 import imp
 import sys
@@ -25,25 +27,48 @@ objects = imp.new_module('objects')
 exec_(OBJECTS_CODE, objects.__dict__, objects.__dict__)
 
 
-@pytest.fixture(scope="module", params=["pure-python", "c-extension"])
-def lazy_object_proxy(request):
+def load_implementation(name):
     class mod:
-        if request.param == "pure-python":
-            from lazy_object_proxy.proxy import Proxy
-        elif request.param == "c-extension":
+        type = name
+        if name == "slots":
+            from lazy_object_proxy.slots import Proxy
+        elif name == "simple":
+            from lazy_object_proxy.simple import Proxy
+        elif name == "cext":
             try:
-                from lazy_object_proxy._proxy import Proxy
+                from lazy_object_proxy.cext import Proxy
             except ImportError:
                 if PYPY:
                     pytest.skip(msg="C Extension not available.")
                 else:
                     raise
+        elif name == "objproxies":
+            Proxy = pytest.importorskip("objproxies").LazyProxy
+        elif name == "django":
+            Proxy = pytest.importorskip("django.utils.functional").SimpleLazyObject
         else:
-            raise RuntimeError("Unsupported param: %r." % request.param)
-
+            raise RuntimeError("Unsupported param: %r." % name)
     return mod
 
 
+@pytest.fixture(scope="module", params=[
+    "slots", "cext",
+    "simple",
+    # "external-django", "external-objproxies"
+])
+def lazy_object_proxy(request):
+    return load_implementation(request.param)
+
+
+def xfail_simple(test_func):
+    from lazy_object_proxy.simple import Proxy
+    @wraps(test_func)
+    def xfail_wrapper(lazy_object_proxy):
+        if lazy_object_proxy.Proxy is Proxy:
+            pytest.xfail(reason="The lazy_object_proxy.simple.Proxy has some limitations.")
+    return xfail_wrapper
+
+
 def test_attributes(lazy_object_proxy):
     def function1(*args, **kwargs):
         return args, kwargs
@@ -73,7 +98,7 @@ def test_set_wrapped(lazy_object_proxy):
     function2 = lazy_object_proxy.Proxy(lambda: function1)
 
     assert function2 == function1
-    assert function2.__wrapped__ == function1
+    assert function2.__wrapped__ is function1
     assert function2.__name__ == function1.__name__
 
     if PY3:
@@ -84,7 +109,7 @@ def test_set_wrapped(lazy_object_proxy):
     assert not hasattr(function1, '__wrapped__')
 
     assert function2 == None
-    assert function2.__wrapped__ == None
+    assert function2.__wrapped__ is None
     assert not hasattr(function2, '__name__')
 
     if PY3:
@@ -103,39 +128,6 @@ def test_set_wrapped(lazy_object_proxy):
         assert function2.__qualname__ == function3.__qualname__
 
 
-def test_delete_wrapped(lazy_object_proxy):
-    def function1(*args, **kwargs):
-        return args, kwargs
-
-    function2 = lazy_object_proxy.Proxy(lambda: function1)
-
-    def run(*args):
-        del function2.__wrapped__
-
-    pytest.raises(TypeError, run, ())
-
-
-def test_proxy_attribute(lazy_object_proxy):
-    def function1(*args, **kwargs):
-        return args, kwargs
-
-    function2 = lazy_object_proxy.Proxy(lambda: function1)
-
-    function2._self_variable = True
-
-    assert not hasattr(function1, '_self_variable')
-    assert hasattr(function2, '_self_variable')
-
-    assert function2._self_variable == True
-
-    del function2._self_variable
-
-    assert not hasattr(function1, '_self_variable')
-    assert not hasattr(function2, '_self_variable')
-
-    assert getattr(function2, '_self_variable', None) == None
-
-
 def test_wrapped_attribute(lazy_object_proxy):
     def function1(*args, **kwargs):
         return args, kwargs
@@ -299,7 +291,7 @@ def test_dir_of_class(lazy_object_proxy):
 
     assert dir(wrapper) == dir(target)
 
-
+@xfail_simple
 def test_vars_of_class(lazy_object_proxy):
     # Test preservation of class __dir__ attribute.
 
@@ -318,6 +310,7 @@ def test_dir_of_instance(lazy_object_proxy):
     assert dir(wrapper) == dir(target)
 
 
+@xfail_simple
 def test_vars_of_instance(lazy_object_proxy):
     # Test preservation of instance __dir__ attribute.
 
@@ -336,6 +329,7 @@ def test_dir_of_function(lazy_object_proxy):
     assert dir(wrapper) == dir(target)
 
 
+@xfail_simple
 def test_vars_of_function(lazy_object_proxy):
     # Test preservation of function __dir__ attribute.
 
@@ -958,21 +952,21 @@ def test_pow(lazy_object_proxy):
     two = lazy_object_proxy.Proxy(lambda: 2)
     three = lazy_object_proxy.Proxy(lambda: 3)
 
-    assert three ** two, pow(3 == 2)
-    assert 3 ** two, pow(3 == 2)
-    assert three ** 2, pow(3 == 2)
+    assert three ** two == pow(3, 2)
+    assert 3 ** two == pow(3, 2)
+    assert three ** 2 == pow(3, 2)
 
-    assert pow(three, two), pow(3 == 2)
-    assert pow(3, two), pow(3 == 2)
-    assert pow(three, 2), pow(3 == 2)
+    assert pow(three, two) == pow(3, 2)
+    assert pow(3, two) == pow(3, 2)
+    assert pow(three, 2) == pow(3, 2)
 
     # Only PyPy implements __rpow__ for ternary pow().
 
     if PYPY:
-        assert pow(three, two, 2), pow(3, 2 == 2)
-        assert pow(3, two, 2), pow(3, 2 == 2)
+        assert pow(three, two, 2) == pow(3, 2, 2)
+        assert pow(3, two, 2) == pow(3, 2, 2)
 
-    assert pow(three, 2, 2), pow(3, 2 == 2)
+    assert pow(three, 2, 2) == pow(3, 2, 2)
 
 
 def test_lshift(lazy_object_proxy):
@@ -1027,12 +1021,14 @@ def test_iadd(lazy_object_proxy):
     value += 1
     assert value == 2
 
-    assert type(value) == lazy_object_proxy.Proxy
+    if lazy_object_proxy.type != 'simple':
+        assert type(value) == lazy_object_proxy.Proxy
 
     value += one
     assert value == 3
 
-    assert type(value) == lazy_object_proxy.Proxy
+    if lazy_object_proxy.type != 'simple':
+        assert type(value) == lazy_object_proxy.Proxy
 
 
 def test_isub(lazy_object_proxy):
@@ -1041,13 +1037,14 @@ def test_isub(lazy_object_proxy):
 
     value -= 1
     assert value == 0
-
-    assert type(value) == lazy_object_proxy.Proxy
+    if lazy_object_proxy.type != 'simple':
+        assert type(value) == lazy_object_proxy.Proxy
 
     value -= one
     assert value == -1
 
-    assert type(value) == lazy_object_proxy.Proxy
+    if lazy_object_proxy.type != 'simple':
+        assert type(value) == lazy_object_proxy.Proxy
 
 
 def test_imul(lazy_object_proxy):
@@ -1057,12 +1054,14 @@ def test_imul(lazy_object_proxy):
     value *= 2
     assert value == 4
 
-    assert type(value) == lazy_object_proxy.Proxy
+    if lazy_object_proxy.type != 'simple':
+        assert type(value) == lazy_object_proxy.Proxy
 
     value *= two
     assert value == 8
 
-    assert type(value) == lazy_object_proxy.Proxy
+    if lazy_object_proxy.type != 'simple':
+        assert type(value) == lazy_object_proxy.Proxy
 
 
 def test_idiv(lazy_object_proxy):
@@ -1075,12 +1074,14 @@ def test_idiv(lazy_object_proxy):
     value /= 2
     assert value == 2 / 2
 
-    assert type(value) == lazy_object_proxy.Proxy
+    if lazy_object_proxy.type != 'simple':
+        assert type(value) == lazy_object_proxy.Proxy
 
     value /= two
     assert value == 2 / 2 / 2
 
-    assert type(value) == lazy_object_proxy.Proxy
+    if lazy_object_proxy.type != 'simple':
+        assert type(value) == lazy_object_proxy.Proxy
 
 
 def test_ifloordiv(lazy_object_proxy):
@@ -1090,12 +1091,14 @@ def test_ifloordiv(lazy_object_proxy):
     value //= 2
     assert value == 2 // 2
 
-    assert type(value) == lazy_object_proxy.Proxy
+    if lazy_object_proxy.type != 'simple':
+        assert type(value) == lazy_object_proxy.Proxy
 
     value //= two
     assert value == 2 // 2 // 2
 
-    assert type(value) == lazy_object_proxy.Proxy
+    if lazy_object_proxy.type != 'simple':
+        assert type(value) == lazy_object_proxy.Proxy
 
 
 def test_imod(lazy_object_proxy):
@@ -1105,12 +1108,14 @@ def test_imod(lazy_object_proxy):
     value %= 2
     assert value == 10 % 2
 
-    assert type(value) == lazy_object_proxy.Proxy
+    if lazy_object_proxy.type != 'simple':
+        assert type(value) == lazy_object_proxy.Proxy
 
     value %= two
     assert value == 10 % 2 % 2
 
-    assert type(value) == lazy_object_proxy.Proxy
+    if lazy_object_proxy.type != 'simple':
+        assert type(value) == lazy_object_proxy.Proxy
 
 
 def test_ipow(lazy_object_proxy):
@@ -1120,12 +1125,14 @@ def test_ipow(lazy_object_proxy):
     value **= 2
     assert value == 10 ** 2
 
-    assert type(value) == lazy_object_proxy.Proxy
+    if lazy_object_proxy.type != 'simple':
+        assert type(value) == lazy_object_proxy.Proxy
 
     value **= two
     assert value == 10 ** 2 ** 2
 
-    assert type(value) == lazy_object_proxy.Proxy
+    if lazy_object_proxy.type != 'simple':
+        assert type(value) == lazy_object_proxy.Proxy
 
 
 def test_ilshift(lazy_object_proxy):
@@ -1135,12 +1142,14 @@ def test_ilshift(lazy_object_proxy):
     value <<= 2
     assert value == 256 << 2
 
-    assert type(value) == lazy_object_proxy.Proxy
+    if lazy_object_proxy.type != 'simple':
+        assert type(value) == lazy_object_proxy.Proxy
 
     value <<= two
     assert value == 256 << 2 << 2
 
-    assert type(value) == lazy_object_proxy.Proxy
+    if lazy_object_proxy.type != 'simple':
+        assert type(value) == lazy_object_proxy.Proxy
 
 
 def test_irshift(lazy_object_proxy):
@@ -1150,12 +1159,14 @@ def test_irshift(lazy_object_proxy):
     value >>= 2
     assert value == 2 >> 2
 
-    assert type(value) == lazy_object_proxy.Proxy
+    if lazy_object_proxy.type != 'simple':
+        assert type(value) == lazy_object_proxy.Proxy
 
     value >>= two
     assert value == 2 >> 2 >> 2
 
-    assert type(value) == lazy_object_proxy.Proxy
+    if lazy_object_proxy.type != 'simple':
+        assert type(value) == lazy_object_proxy.Proxy
 
 
 def test_iand(lazy_object_proxy):
@@ -1165,12 +1176,14 @@ def test_iand(lazy_object_proxy):
     value &= 2
     assert value == 1 & 2
 
-    assert type(value) == lazy_object_proxy.Proxy
+    if lazy_object_proxy.type != 'simple':
+        assert type(value) == lazy_object_proxy.Proxy
 
     value &= two
     assert value == 1 & 2 & 2
 
-    assert type(value) == lazy_object_proxy.Proxy
+    if lazy_object_proxy.type != 'simple':
+        assert type(value) == lazy_object_proxy.Proxy
 
 
 def test_ixor(lazy_object_proxy):
@@ -1180,12 +1193,14 @@ def test_ixor(lazy_object_proxy):
     value ^= 2
     assert value == 1 ^ 2
 
-    assert type(value) == lazy_object_proxy.Proxy
+    if lazy_object_proxy.type != 'simple':
+        assert type(value) == lazy_object_proxy.Proxy
 
     value ^= two
     assert value == 1 ^ 2 ^ 2
 
-    assert type(value) == lazy_object_proxy.Proxy
+    if lazy_object_proxy.type != 'simple':
+        assert type(value) == lazy_object_proxy.Proxy
 
 
 def test_ior(lazy_object_proxy):
@@ -1195,12 +1210,14 @@ def test_ior(lazy_object_proxy):
     value |= 2
     assert value == 1 | 2
 
-    assert type(value) == lazy_object_proxy.Proxy
+    if lazy_object_proxy.type != 'simple':
+        assert type(value) == lazy_object_proxy.Proxy
 
     value |= two
     assert value == 1 | 2 | 2
 
-    assert type(value) == lazy_object_proxy.Proxy
+    if lazy_object_proxy.type != 'simple':
+        assert type(value) == lazy_object_proxy.Proxy
 
 
 def test_neg(lazy_object_proxy):
@@ -1290,7 +1307,7 @@ def test_delitem(lazy_object_proxy):
 def test_getslice(lazy_object_proxy):
     value = lazy_object_proxy.Proxy(lambda: list(range(5)))
 
-    assert value[1:4], [1, 2 == 3]
+    assert value[1:4] == [1, 2, 3]
 
 
 def test_setslice(lazy_object_proxy):
@@ -1298,7 +1315,7 @@ def test_setslice(lazy_object_proxy):
 
     value[1:4] = reversed(value[1:4])
 
-    assert value[1:4], [3, 2 == 1]
+    assert value[1:4] == [3, 2, 1]
 
 
 def test_delslice(lazy_object_proxy):
@@ -1307,7 +1324,7 @@ def test_delslice(lazy_object_proxy):
     del value[1:4]
 
     assert len(value) == 2
-    assert value, [0 == 4]
+    assert value == [0, 4]
 
 
 def test_length(lazy_object_proxy):
@@ -1320,7 +1337,7 @@ def test_contains(lazy_object_proxy):
     value = lazy_object_proxy.Proxy(lambda: dict.fromkeys(range(3), False))
 
     assert 2 in value
-    assert not -2 in value
+    assert -2 not in value
 
 
 def test_getitem(lazy_object_proxy):
@@ -1386,18 +1403,6 @@ def test_derived_new(lazy_object_proxy):
     obj = DerivedObjectProxy(lambda: function)
 
 
-def test_derived_setattr(lazy_object_proxy):
-    class DerivedObjectProxy(lazy_object_proxy.Proxy):
-        def __init__(self, wrapped):
-            self._self_attribute = True
-            super(DerivedObjectProxy, self).__init__(wrapped)
-
-    def function():
-        pass
-
-    obj = DerivedObjectProxy(lambda: function)
-
-
 def test_setup_class_attributes(lazy_object_proxy):
     def function():
         pass
@@ -1419,6 +1424,7 @@ def test_setup_class_attributes(lazy_object_proxy):
     assert not hasattr(function, 'ATTRIBUTE')
 
 
+@xfail_simple
 def test_override_class_attributes(lazy_object_proxy):
     def function():
         pass
@@ -1445,55 +1451,6 @@ def test_override_class_attributes(lazy_object_proxy):
     assert not hasattr(function, 'ATTRIBUTE')
 
 
-def test_class_properties(lazy_object_proxy):
-    def function():
-        pass
-
-    class DerivedObjectProxy(lazy_object_proxy.Proxy):
-        def __init__(self, wrapped):
-            super(DerivedObjectProxy, self).__init__(wrapped)
-            self._self_attribute = 1
-
-        @property
-        def ATTRIBUTE(self):
-            return self._self_attribute
-
-        @ATTRIBUTE.setter
-        def ATTRIBUTE(self, value):
-            self._self_attribute = value
-
-        @ATTRIBUTE.deleter
-        def ATTRIBUTE(self):
-            del self._self_attribute
-
-    obj = DerivedObjectProxy(lambda: function)
-
-    assert obj.ATTRIBUTE == 1
-
-    obj.ATTRIBUTE = 2
-
-    assert obj.ATTRIBUTE == 2
-    assert not hasattr(function, 'ATTRIBUTE')
-
-    del obj.ATTRIBUTE
-
-    assert not hasattr(obj, 'ATTRIBUTE')
-    assert not hasattr(function, 'ATTRIBUTE')
-
-    obj.ATTRIBUTE = 1
-
-    assert obj.ATTRIBUTE == 1
-
-    obj.ATTRIBUTE = 2
-
-    assert obj.ATTRIBUTE == 2
-    assert not hasattr(function, 'ATTRIBUTE')
-
-    del obj.ATTRIBUTE
-
-    assert not hasattr(obj, 'ATTRIBUTE')
-    assert not hasattr(function, 'ATTRIBUTE')
-
 
 def test_attr_functions(lazy_object_proxy):
     def function():
@@ -1555,7 +1512,6 @@ def test_proxy_is_callable(lazy_object_proxy):
     assert not callable(proxy)
 
 
-@skipcallable
 def test_callable_proxy_hasattr_call(lazy_object_proxy):
     proxy = lazy_object_proxy.Proxy(lambda: None)
 
@@ -1569,7 +1525,6 @@ def test_callable_proxy_getattr_call(lazy_object_proxy):
     assert getattr(proxy, '__call__', None) is None
 
 
-@skipcallable
 def test_callable_proxy_is_callable(lazy_object_proxy):
     proxy = lazy_object_proxy.Proxy(lambda: None)
 
@@ -1625,11 +1580,94 @@ def test_fractions_round(lazy_object_proxy):
     assert round(instance) == round(proxy)
 
 
+def test_readonly(lazy_object_proxy):
+    class Foo(object):
+        if PY2:
+            @property
+            def __qualname__(self):
+                return 'object'
+    proxy = lazy_object_proxy.Proxy(lambda: Foo() if PY2 else object)
+    assert proxy.__qualname__ == 'object'
+
+
+def test_del_wrapped(lazy_object_proxy):
+    class Foo(object):
+        pass
+    proxy = lazy_object_proxy.Proxy(lambda: Foo)
+    str(proxy)
+    assert proxy.__wrapped__
+    # print(type(proxy), hasattr(type(proxy), '__wrapped__'))
+    del proxy.__wrapped__
+
+
 def test_new(lazy_object_proxy):
     a = lazy_object_proxy.Proxy.__new__(lazy_object_proxy.Proxy)
     b = lazy_object_proxy.Proxy.__new__(lazy_object_proxy.Proxy)
     # NOW KISS
-    pytest.raises(ValueError, lambda: a.__wrapped__)
     pytest.raises(ValueError, lambda: a + b)
     # no segfault, yay
+    pytest.raises(ValueError, lambda: a.__wrapped__)
 
+
+@pytest.mark.parametrize("name", ["slots", "cext", "simple", "django", "objproxies"])
+def test_perf(benchmark, name):
+    implementation = load_implementation(name)
+    obj = "foobar"
+    proxied = implementation.Proxy(lambda: obj)
+    assert benchmark(partial(str, proxied)) == obj
+
+empty = object()
+
+
+@pytest.fixture(scope="module", params=["SimpleProxy", "LocalsSimpleProxy", "CachedPropertyProxy", "LocalsCachedPropertyProxy"])
+def prototype(request):
+    from lazy_object_proxy.simple import cached_property
+    name = request.param
+
+    if name == "SimpleProxy":
+        class SimpleProxy(object):
+            def __init__(self, factory):
+                self.factory = factory
+                self.object = empty
+            def __str__(self):
+                if self.object is empty:
+                    self.object = self.factory()
+                return str(self.object)
+        return SimpleProxy
+    elif name == "CachedPropertyProxy":
+        class CachedPropertyProxy(object):
+            def __init__(self, factory):
+                self.factory = factory
+            @cached_property
+            def object(self):
+                return self.factory()
+            def __str__(self):
+                return str(self.object)
+        return CachedPropertyProxy
+    elif name == "LocalsSimpleProxy":
+        class LocalsSimpleProxy(object):
+            def __init__(self, factory):
+                self.factory = factory
+                self.object = empty
+            def __str__(self, func=str):
+                if self.object is empty:
+                    self.object = self.factory()
+                return func(self.object)
+        return LocalsSimpleProxy
+    elif name == "LocalsCachedPropertyProxy":
+        class LocalsCachedPropertyProxy(object):
+            def __init__(self, factory):
+                self.factory = factory
+            @cached_property
+            def object(self):
+                return self.factory()
+            def __str__(self, func=str):
+                return func(self.object)
+        return LocalsCachedPropertyProxy
+
+
+@pytest.mark.benchmark(group="prototypes")
+def test_proto(benchmark, prototype):
+    obj = "foobar"
+    proxied = prototype(lambda: obj)
+    assert benchmark(partial(str, proxied)) == obj
